/*--------------------------------------------------------------------------*/
/*  PLOT_ANSI: package for simple 2d X11 graphics                           */
/*                                                                          */
/*  (c) by A. Schmidt (1990-2004)                                           */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <stdio.h>

#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/* single of double buffered ? */
#define single 




#define PLOT_EXTERN   /* plot externals are defined HERE! */
#include "PLOTansi.h"


PLOT_INFO DUMMY_PLOT_INFO  = {NULL, NULL, NULL, 0, NULL, NULL};

PLOT_INFO *current_PLOT_info = &DUMMY_PLOT_INFO;


#define INFO current_PLOT_info





/*PLOT_ErrorHandler is only used in Xdirect mode */ 

#ifdef XDIRECT

static int PLOT_ErrorHandler(int arg1, int arg2)
{
        XdReset();  /* ok, because only used in Xdirect mode */

/*        printf("PLOT_ErrorHandler(%d,%d) -- Reset Xdirect\n",arg1,arg2); */
	fflush(stdout);
	PLOT_redraw = 1;
	
/*	PLOT_expose(NULL); */

        return(0);
}
#endif

/**********************************************************************
  WindowHandle *PLOT_create_window(name,geometry)
       char *name,*geometry;

  create a new current PLOT window with given name and default geometry
***********************************************************************/

PLOT_INFO *PLOT_create_window(const char * name, const char *geometry)
{
  Window plot_window;
  XSizeHints size_hints;
  Colormap colourmap;
  int screen,status;
  char *displayname = NULL;
  static char *default_geometry = "=500x500+0+0";
  int borderwidth = 2;
  XColor borderColor, backgrColor;
  char *borderColorName, *backgrColorName;
  static Display *display=NULL;
  WindowHandle *wh;
  XEvent report;
  PLOT_INFO *info;
  static float Black[3] = {0.0, 0.0, 0.0};
  static float White[3] = {1.0, 1.0, 1.0};
#if 0
  static float Blue[3]  = {0.0, 0.0, 1.0};
  static float Green[3] = {0.0, 1.0, 0.0};
#endif
  int i,j;
#ifdef XDIRECT
  int buffer = 0;
  char *display_string;
  static char *direct_string="titan";
#endif


  if (display==NULL) {
/*    printf("Versuche, X zu erreichen\n"); */
    if ((display = XOpenDisplay(displayname))==NULL) {
      printf(" cannot connect to display.\n");
      return(NULL);
    }
  }

  screen=DefaultScreen(display);



  size_hints.flags = PResizeInc | PMinSize ;
  size_hints.width_inc  =  1;
  size_hints.height_inc =  1;
  size_hints.min_width  = 10;
  size_hints.min_height = 10;

  if (geometry) {
    status = XGeometry(display, screen, geometry, default_geometry,
		       borderwidth, 1, 1, 0, 0, 
		       &(size_hints.x), &(size_hints.y),
		       &(size_hints.width), &(size_hints.height));
    if (status & (XValue|YValue)) {
      size_hints.flags |= USPosition;
      size_hints.flags &= ~PPosition;
    }
    if (status & (WidthValue|HeightValue)) {
      size_hints.flags |= USSize;
      size_hints.flags &= ~PSize;
    }
  }

  if (size_hints.flags & USSize) {
/*    printf(" set aspect ratio\n"); */
    size_hints.flags |= PAspect;
    size_hints.min_aspect.x = size_hints.width;
    size_hints.min_aspect.y = size_hints.height;
    size_hints.max_aspect.x = size_hints.width;
    size_hints.max_aspect.y = size_hints.height;
  }


  

  /*colourmap=XTitanDefaultDirectColormap(display,screen);*/
  colourmap=DefaultColormap(display,screen);
  

  
  if ((borderColorName = XGetDefault(display, name, "borderColor")));
  else if ((borderColorName = XGetDefault(display, "PLOT", "borderColor")));
  else borderColorName = "Blue";
/*  printf("borderColorName='%s'\n",borderColorName); */
  if (XParseColor(display, colourmap, borderColorName, &borderColor));
  else XParseColor(display, colourmap, "Black", &borderColor);
  XAllocColor(display, colourmap, &borderColor);
  
  
  if ((backgrColorName = XGetDefault(display, name, "Background")));
  else if ((backgrColorName = XGetDefault(display, "PLOT", "Background")));
  else backgrColorName = "Green";
/*  printf("backgrColorName='%s'\n",backgrColorName); */
  if (XParseColor(display, colourmap, backgrColorName, &backgrColor));
  else XParseColor(display, colourmap, "White", &backgrColor);
  XAllocColor(display, colourmap, &backgrColor);
  

  
  /*(display,parent,x,y,breite,hoehe,randbreite,tiefe,classe,visual,
     wertmaske[zu attributen gehoerig],attribute) */

  plot_window = XCreateSimpleWindow(display, 
				    XDefaultRootWindow(display),
				    size_hints.x, size_hints.y,
				    size_hints.width, size_hints.height,
				    2, 
				    borderColor.pixel, backgrColor.pixel );

  if (plot_window == 0) { /* plot_window is the window ID, not a pointer */
    printf("cannot create plot_window!");
    return(NULL);
  }


  XSetStandardProperties(display, plot_window,
			 name, name, None, NULL, 0, &size_hints);

/*  printf("select input for window %d\n",plot_window); */

  XSelectInput(display, plot_window,
 	       ExposureMask | KeyPressMask 
	       | ButtonPressMask 
/*	       | StructureNotifyMask  */  /* gives problems! */
	       );

  XMapWindow(display, plot_window);

/* wait for Exposure event resulting from XMapWindow */

  XWindowEvent(display, plot_window, ExposureMask, &report);
  while (XCheckWindowEvent(display, plot_window, ExposureMask, &report)) ;


#ifdef XDIRECT
  display_string = DisplayString(display);
  if (strncmp(display_string, direct_string, strlen(direct_string)) == 0) {
    /*  printf("Versuche, Fenster fuer Grafik zu oeffnen\n");*/
    wh = 0;
 /*   wh=XdCreateWindowHandle(display,plot_window); */

    if(wh) {
      /*    printf("Habe Handle bekommen: %d\n",wh); */
    }
    else {
      printf("Habe KEINEN Handle bekommen; Standard X11 Display.\n");
    }
  }
  else {
    wh = 0;
    /* printf("generate standard X11 output\n"); */
  }
#else
    wh = 0;
#endif

  
  info = (PLOT_INFO *)MALLOC(sizeof(PLOT_INFO));
  if (info) {
    info->prev = INFO;
    info->next = INFO->next;
    if (info->next) info->next->prev = info;
    INFO->next = info;
    INFO = info;
  }
  else {
    printf("couldn't malloc a PLOT_INFO structure.\n");
    return(NULL);
  }

  INFO->display = display;
  INFO->window = plot_window;
  INFO->wh = wh;
  INFO->windowWidth = 0;
  INFO->windowHeight = 0;
  INFO->log = NULL;
  INFO->current_color = 0;
  for (i=0;i<3;i++) INFO->background_color[i] = White[i];
  for (i=0;i<3;i++) INFO->line_color[i] = Black[i];
  for (i=0;i<3;i++) INFO->patch_color[i] = Black[i];
  for (i=0;i<3;i++) INFO->text_color[i] = Black[i];
  for(i=0;i<3;i++) for(j=0;j<3;j++) 
    INFO->model_matrix  [i][j] = (i==j? 1.0 : 0.0);
  for(i=0;i<3;i++) for(j=0;j<3;j++) 
    INFO->view_matrix   [i][j] = (i==j? 1.0 : 0.0);
  for(i=0;i<3;i++) for(j=0;j<3;j++) 
    INFO->project_matrix[i][j] = (i==j? 1.0 : 0.0);
  for(i=0;i<3;i++) for(j=0;j<3;j++) 
    INFO->W2W_matrix    [i][j] = (i==j? 1.0 : 0.0);

  XSetWindowBackground(display, plot_window, WhitePixel(display,screen));

  INFO->gc = XCreateGC(display, RootWindow(display, screen), 0, NULL);
  XSetBackground(display, INFO->gc, WhitePixel(display,screen));
  XSetForeground(display, INFO->gc, BlackPixel(display,screen));

  PLOT_redraw = 0;

/*  printf("Versuche zu zeichnen\n"); */
  
#ifdef XDIRECT
  if (wh) {
/*    XdSetZFunction(wh,Z_FUNC_GE);  */
    XdSetDrawBuffer(wh,buffer,0);
    XdSetErrorHandler(PLOT_ErrorHandler);
  }
#endif

  PLOT_expose(plot_window);
  
  return(INFO);

}



/**********************************************************************
  void PLOT_destroy_window()

  destroy a PLOT window
***********************************************************************/

void PLOT_destroy_window(PLOT_INFO *info)
{
  if (info) {
    if (info == INFO) {
      if (info->next) INFO = info->next;
      else if (info->prev) INFO = info->prev;
      else INFO = &DUMMY_PLOT_INFO;
    }
    if (info->prev) info->prev->next = info->next;
    if (info->next) info->next->prev = info->prev;

    if (info->wh) {
#ifdef XDIRECT
/********************************************** BUG FIX ? ***********
      XdDestroyWindowHandle(info->wh); 
      info->wh = 0;
      if (info->display && info->window) 
        XDestroyWindow(info->display, info->window); *******************/
/* preceeding code makes REAL problems ( when one window-handle is destroyed
   and another one exists longer, and XdReset() is called. 
   probably a bug in XdReset().   
   So...
 */
      XUnmapWindow(info->display, info->window); 
/***********************************************************************/
#endif
    }
    else {
      XDestroyWindow(info->display, info->window);
    }


    free_log(info->log);
    info->log = NULL;

    MFREE(info, sizeof(PLOT_INFO));
  }
}



/**********************************************************************
  PLOT_INFO *PLOT_get_std_window()

  return PLOT_INFO of current PLOT window
***********************************************************************/

PLOT_INFO *PLOT_get_std_window(void)
{
  return(INFO);
}



/**********************************************************************
  PLOT_INFO *PLOT_set_std_window(new_window)
       PLOT_INFO *new_window;

  set current PLOT window to new_window
  return PLOT_INFO of previously current PLOT window
***********************************************************************/

PLOT_INFO *PLOT_set_std_window(PLOT_INFO *new_window)
{
  PLOT_INFO *old_window;

  old_window = INFO;
  if (new_window) {  /* change stdwindow only if new one exists */
    INFO = new_window;
  }
  return(old_window);
}





/**********************************************************************
  PLOT_expose(window)
       Window window;

  recalculate window size after an expose
***********************************************************************/

void PLOT_expose(Window window)
{
  int i;
  PLOT_INFO *info;
  PLOT_LOG *log;
  XWindowAttributes attis;

 again:
  PLOT_redraw = 0;

  info = INFO;

#ifdef XDIRECT
  if (info->wh) {
    XdReset();
  }
#endif

  for (i=0;(i<100) && (info->prev);i++)
    info = info->prev;  /* hope there is no endless loop */

  if (info->prev) {
    printf(" PLOT_expose: endless loop (100)!\n");
    exit(0);
  }

  while (info) {
    if (PLOT_redraw) {
      printf("again PLOT_expose 2\n");
      goto again;
    }
    if ((window==0) || (window==info->window)) { /* window is an integer, not
						  * a pointer
						  */

      /*printf("PLOT_expose window %d\n",info->window);*/
      if (info->window) {
	XGetWindowAttributes( info->display, info->window, &attis );
	info->windowWidth  = attis.width;
	info->windowHeight = attis.height;
	/* set_W2W_matrix2d(info); */

	/* reset info->log->flushed */
	log = info->log;
	for (i=0;(i<10000) && log; i++) {
	  log->flushed = 0;
	  log = log->prev;
	}

	/*printf("PLOT_expose calls PLOT_LOG_flush\n");*/
	PLOT_LOG_flush(info);
      }
    }
    info = info->next;
  }

  if (PLOT_redraw) {
/*    printf("again PLOT_expose 2\n"); */
    goto again;
  }

}




/**********************************************************************
  PLOT_title(title)
       char *title

  set title of current PLOT window
***********************************************************************/

void PLOT_title(char *title)
{
  if (INFO && INFO->display && INFO->window ) {
/*    printf(" new window title for display %d, window %d\n",
	   INFO->display, INFO->window);
*/
    XStoreName(INFO->display, INFO->window, title);
  }
}



/**********************************************************************
  free_log(log)
       PLOT_LOG *log;

  free a whole list of PLOT_LOG structures
***********************************************************************/

void free_log(PLOT_LOG *log)
{
  int i;
  PLOT_LOG *this, *next;


  if (log) {
    this = log;
    for (i=0;i<10000;i++) {
      if (this->prev) {
	this = this->prev;
      }
      else break;
    }
    if (this->prev) {
      printf("free_log: endless loop 1.\n");
      return;
    }

    for (i=0;i<10000;i++) {
      if (this==NULL) break;
      next = this->next;
      MFREE(this, sizeof(PLOT_LOG));
      /* printf(" free log %d\n",this); */
      this = next;
    }
    if (this) {
      printf("free_log: endless loop 2.\n");
      return;
    }

  }
}



/**********************************************************************
  PLOT_alarm(sec)
       unsigned sec;

  start the alarm window handler with interval of <sec> seconds
***********************************************************************/

#include <signal.h>

static unsigned PLOT_ALARM_seconds = 0;


void PLOT_alarm(unsigned sec)
{

  if (sec > 10)  sec = 10;

  PLOT_ALARM_seconds = sec;

  if (sec) {
    signal(SIGALRM, PLOT_ALARM_HANDLER);
    alarm(sec);
    /*printf(" PLOT alarm handler started with interval %d sec\n",sec);*/
  }
  else {
    signal(SIGALRM, SIG_IGN);
    alarm(0);
    /*printf(" PLOT alarm handler stopped\n");*/
  }
}


void PLOT_ALARM_HANDLER(int i)
{
  XEvent report;
  Window window;

  /* printf("PLOT_ALARM_HANDLER()\n"); */

  extern int PLOT_LOG_flushing;

  if (PLOT_LOG_flushing==0) {
    if (INFO && INFO->display) {
      while (XCheckTypedEvent(INFO->display, Expose, &report)==True) {
	window = report.xexpose.window;
	/* printf("expose event window=%d\n",window); */
	while (XCheckWindowEvent(INFO->display, window, Expose, &report)) ;
	/*printf(" alarm calls PLOT_expose\n");*/
	PLOT_expose(window);
      }
    }
  }
  else {
    /*printf(" alarm while flushing ; wait now...\n");*/
  }

  if (PLOT_ALARM_seconds) {
    signal(SIGALRM, PLOT_ALARM_HANDLER);
    alarm(PLOT_ALARM_seconds);
  }
  /* printf("restart alarm, %d seconds.\n",PLOT_ALARM_seconds); */
}
